@@ -2,14 +2,13 @@
 
   cont.c -
 
-  $Author: akr $
+  $Author: usa $
   created at: Thu May 23 09:03:43 2007
 
   Copyright (C) 2007 Koichi Sasada
 
 **********************************************************************/
 
-#include "ruby/ruby.h"
 #include "internal.h"
 #include "vm_core.h"
 #include "gc.h"
@@ -89,24 +88,26 @@ enum context_type {
 
 typedef struct rb_context_struct {
     enum context_type type;
-    VALUE self;
     int argc;
+    VALUE self;
     VALUE value;
     VALUE *vm_stack;
 #ifdef CAPTURE_JUST_VALID_VM_STACK
     size_t vm_stack_slen;  /* length of stack (head of th->stack) */
     size_t vm_stack_clen;  /* length of control frames (tail of th->stack) */
 #endif
-    VALUE *machine_stack;
-    VALUE *machine_stack_src;
+    struct {
+	VALUE *stack;
+	VALUE *stack_src;
+	size_t stack_size;
 #ifdef __ia64
-    VALUE *machine_register_stack;
-    VALUE *machine_register_stack_src;
-    int machine_register_stack_size;
+	VALUE *register_stack;
+	VALUE *register_stack_src;
+	int register_stack_size;
 #endif
-    rb_thread_t saved_thread;
+    } machine;
+    rb_thread_t saved_thread; /* selected properties of GET_THREAD() (see cont_save_thread) */
     rb_jmpbuf_t jmpbuf;
-    size_t machine_stack_size;
     rb_ensure_entry_t *ensure_array;
     rb_ensure_list_t *ensure_list;
 } rb_context_t;
@@ -128,26 +129,30 @@ static machine_stack_cache_t machine_stack_cache[MAX_MACHINE_STACK_CACHE];
 static machine_stack_cache_t terminated_machine_stack;
 #endif
 
-typedef struct rb_fiber_struct {
+struct rb_fiber_struct {
     rb_context_t cont;
-    VALUE prev;
+    struct rb_fiber_struct *prev;
     enum fiber_status status;
-    struct rb_fiber_struct *prev_fiber;
-    struct rb_fiber_struct *next_fiber;
     /* If a fiber invokes "transfer",
      * then this fiber can't "resume" any more after that.
      * You shouldn't mix "transfer" and "resume".
      */
-    int transfered;
+    int transferred;
 
 #if FIBER_USE_NATIVE
 #ifdef _WIN32
     void *fib_handle;
 #else
     ucontext_t context;
+    /* Because context.uc_stack.ss_sp and context.uc_stack.ss_size
+     * are not necessarily valid after makecontext() or swapcontext(),
+     * they are saved in these variables for later use.
+     */
+    void *ss_sp;
+    size_t ss_size;
 #endif
 #endif
-} rb_fiber_t;
+};
 
 static const rb_data_type_t cont_data_type, fiber_data_type;
 static VALUE rb_cContinuation;
@@ -183,32 +188,32 @@ cont_mark(void *ptr)
 	    rb_gc_mark_locations(cont->vm_stack,
 				 cont->vm_stack + cont->vm_stack_slen + cont->vm_stack_clen);
 #else
-	    rb_gc_mark_localtion(cont->vm_stack,
+	    rb_gc_mark_locations(cont->vm_stack,
 				 cont->vm_stack, cont->saved_thread.stack_size);
 #endif
 	}
 
-	if (cont->machine_stack) {
+	if (cont->machine.stack) {
 	    if (cont->type == CONTINUATION_CONTEXT) {
 		/* cont */
-		rb_gc_mark_locations(cont->machine_stack,
-				     cont->machine_stack + cont->machine_stack_size);
+		rb_gc_mark_locations(cont->machine.stack,
+				     cont->machine.stack + cont->machine.stack_size);
             }
             else {
 		/* fiber */
 		rb_thread_t *th;
                 rb_fiber_t *fib = (rb_fiber_t*)cont;
 		GetThreadPtr(cont->saved_thread.self, th);
-		if ((th->fiber != cont->self) && fib->status == RUNNING) {
-		    rb_gc_mark_locations(cont->machine_stack,
-					 cont->machine_stack + cont->machine_stack_size);
+		if ((th->fiber != fib) && fib->status == RUNNING) {
+		    rb_gc_mark_locations(cont->machine.stack,
+					 cont->machine.stack + cont->machine.stack_size);
 		}
 	    }
 	}
 #ifdef __ia64
-	if (cont->machine_register_stack) {
-	    rb_gc_mark_locations(cont->machine_register_stack,
-				 cont->machine_register_stack + cont->machine_register_stack_size);
+	if (cont->machine.register_stack) {
+	    rb_gc_mark_locations(cont->machine.register_stack,
+				 cont->machine.register_stack + cont->machine.register_stack_size);
 	}
 #endif
     }
@@ -226,12 +231,13 @@ cont_free(void *ptr)
 	if (cont->type == CONTINUATION_CONTEXT) {
 	    /* cont */
 	    ruby_xfree(cont->ensure_array);
-	    RUBY_FREE_UNLESS_NULL(cont->machine_stack);
+	    RUBY_FREE_UNLESS_NULL(cont->machine.stack);
 	}
 	else {
 	    /* fiber */
+	    rb_fiber_t *fib = (rb_fiber_t*)cont;
 #ifdef _WIN32
-	    if (GET_THREAD()->fiber != cont->self && cont->type != ROOT_FIBER_CONTEXT) {
+	    if (GET_THREAD()->fiber != fib && cont->type != ROOT_FIBER_CONTEXT) {
 		/* don't delete root fiber handle */
 		rb_fiber_t *fib = (rb_fiber_t*)cont;
 		if (fib->fib_handle) {
@@ -239,13 +245,13 @@ cont_free(void *ptr)
 		}
 	    }
 #else /* not WIN32 */
-	    if (GET_THREAD()->fiber != cont->self) {
+	    if (GET_THREAD()->fiber != fib) {
                 rb_fiber_t *fib = (rb_fiber_t*)cont;
-                if (fib->context.uc_stack.ss_sp) {
+                if (fib->ss_sp) {
                     if (cont->type == ROOT_FIBER_CONTEXT) {
 			rb_bug("Illegal root fiber parameter");
                     }
-		    munmap((void*)fib->context.uc_stack.ss_sp, fib->context.uc_stack.ss_size);
+		    munmap((void*)fib->ss_sp, fib->ss_size);
 		}
 	    }
             else {
@@ -257,10 +263,10 @@ cont_free(void *ptr)
 	}
 #else /* not FIBER_USE_NATIVE */
 	ruby_xfree(cont->ensure_array);
-	RUBY_FREE_UNLESS_NULL(cont->machine_stack);
+	RUBY_FREE_UNLESS_NULL(cont->machine.stack);
 #endif
 #ifdef __ia64
-	RUBY_FREE_UNLESS_NULL(cont->machine_register_stack);
+	RUBY_FREE_UNLESS_NULL(cont->machine.register_stack);
 #endif
 	RUBY_FREE_UNLESS_NULL(cont->vm_stack);
 
@@ -286,51 +292,37 @@ cont_memsize(const void *ptr)
 	    size += n * sizeof(*cont->vm_stack);
 	}
 
-	if (cont->machine_stack) {
-	    size += cont->machine_stack_size * sizeof(*cont->machine_stack);
+	if (cont->machine.stack) {
+	    size += cont->machine.stack_size * sizeof(*cont->machine.stack);
 	}
 #ifdef __ia64
-	if (cont->machine_register_stack) {
-	    size += cont->machine_register_stack_size * sizeof(*cont->machine_register_stack);
+	if (cont->machine.register_stack) {
+	    size += cont->machine.register_stack_size * sizeof(*cont->machine.register_stack);
 	}
 #endif
     }
     return size;
 }
 
+void
+rb_fiber_mark_self(rb_fiber_t *fib)
+{
+    if (fib)
+	rb_gc_mark(fib->cont.self);
+}
+
 static void
 fiber_mark(void *ptr)
 {
     RUBY_MARK_ENTER("cont");
     if (ptr) {
 	rb_fiber_t *fib = ptr;
-	rb_gc_mark(fib->prev);
+	rb_fiber_mark_self(fib->prev);
 	cont_mark(&fib->cont);
     }
     RUBY_MARK_LEAVE("cont");
 }
 
-static void
-fiber_link_join(rb_fiber_t *fib)
-{
-    VALUE current_fibval = rb_fiber_current();
-    rb_fiber_t *current_fib;
-    GetFiberPtr(current_fibval, current_fib);
-
-    /* join fiber link */
-    fib->next_fiber = current_fib->next_fiber;
-    fib->prev_fiber = current_fib;
-    current_fib->next_fiber->prev_fiber = fib;
-    current_fib->next_fiber = fib;
-}
-
-static void
-fiber_link_remove(rb_fiber_t *fib)
-{
-    fib->prev_fiber->next_fiber = fib->next_fiber;
-    fib->next_fiber->prev_fiber = fib->prev_fiber;
-}
-
 static void
 fiber_free(void *ptr)
 {
@@ -341,7 +333,6 @@ fiber_free(void *ptr)
 	    fib->cont.saved_thread.local_storage) {
 	    st_free_table(fib->cont.saved_thread.local_storage);
 	}
-	fiber_link_remove(fib);
 
 	cont_free(&fib->cont);
     }
@@ -380,63 +371,82 @@ cont_save_machine_stack(rb_thread_t *th, rb_context_t *cont)
 {
     size_t size;
 
-    SET_MACHINE_STACK_END(&th->machine_stack_end);
+    SET_MACHINE_STACK_END(&th->machine.stack_end);
 #ifdef __ia64
-    th->machine_register_stack_end = rb_ia64_bsp();
+    th->machine.register_stack_end = rb_ia64_bsp();
 #endif
 
-    if (th->machine_stack_start > th->machine_stack_end) {
-	size = cont->machine_stack_size = th->machine_stack_start - th->machine_stack_end;
-	cont->machine_stack_src = th->machine_stack_end;
+    if (th->machine.stack_start > th->machine.stack_end) {
+	size = cont->machine.stack_size = th->machine.stack_start - th->machine.stack_end;
+	cont->machine.stack_src = th->machine.stack_end;
     }
     else {
-	size = cont->machine_stack_size = th->machine_stack_end - th->machine_stack_start;
-	cont->machine_stack_src = th->machine_stack_start;
+	size = cont->machine.stack_size = th->machine.stack_end - th->machine.stack_start;
+	cont->machine.stack_src = th->machine.stack_start;
     }
 
-    if (cont->machine_stack) {
-	REALLOC_N(cont->machine_stack, VALUE, size);
+    if (cont->machine.stack) {
+	REALLOC_N(cont->machine.stack, VALUE, size);
     }
     else {
-	cont->machine_stack = ALLOC_N(VALUE, size);
+	cont->machine.stack = ALLOC_N(VALUE, size);
     }
 
     FLUSH_REGISTER_WINDOWS;
-    MEMCPY(cont->machine_stack, cont->machine_stack_src, VALUE, size);
+    MEMCPY(cont->machine.stack, cont->machine.stack_src, VALUE, size);
 
 #ifdef __ia64
     rb_ia64_flushrs();
-    size = cont->machine_register_stack_size = th->machine_register_stack_end - th->machine_register_stack_start;
-    cont->machine_register_stack_src = th->machine_register_stack_start;
-    if (cont->machine_register_stack) {
-	REALLOC_N(cont->machine_register_stack, VALUE, size);
+    size = cont->machine.register_stack_size = th->machine.register_stack_end - th->machine.register_stack_start;
+    cont->machine.register_stack_src = th->machine.register_stack_start;
+    if (cont->machine.register_stack) {
+	REALLOC_N(cont->machine.register_stack, VALUE, size);
     }
     else {
-	cont->machine_register_stack = ALLOC_N(VALUE, size);
+	cont->machine.register_stack = ALLOC_N(VALUE, size);
     }
 
-    MEMCPY(cont->machine_register_stack, cont->machine_register_stack_src, VALUE, size);
+    MEMCPY(cont->machine.register_stack, cont->machine.register_stack_src, VALUE, size);
 #endif
 }
 
 static const rb_data_type_t cont_data_type = {
     "continuation",
     {cont_mark, cont_free, cont_memsize,},
-    NULL, NULL, RUBY_TYPED_FREE_IMMEDIATELY
+    0, 0, RUBY_TYPED_FREE_IMMEDIATELY
 };
 
-static void
+static inline void
 cont_save_thread(rb_context_t *cont, rb_thread_t *th)
 {
+    rb_thread_t *sth = &cont->saved_thread;
+
     /* save thread context */
-    cont->saved_thread = *th;
-    /* saved_thread->machine_stack_(start|end) should be NULL */
+    sth->stack = th->stack;
+    sth->stack_size = th->stack_size;
+    sth->local_storage = th->local_storage;
+    sth->cfp = th->cfp;
+    sth->safe_level = th->safe_level;
+    sth->raised_flag = th->raised_flag;
+    sth->state = th->state;
+    sth->status = th->status;
+    sth->tag = th->tag;
+    sth->protect_tag = th->protect_tag;
+    sth->errinfo = th->errinfo;
+    sth->first_proc = th->first_proc;
+    sth->root_lep = th->root_lep;
+    sth->root_svar = th->root_svar;
+    sth->ensure_list = th->ensure_list;
+
+    sth->trace_arg = th->trace_arg;
+
+    /* saved_thread->machine.stack_(start|end) should be NULL */
     /* because it may happen GC afterward */
-    cont->saved_thread.machine_stack_start = 0;
-    cont->saved_thread.machine_stack_end = 0;
+    sth->machine.stack_start = 0;
+    sth->machine.stack_end = 0;
 #ifdef __ia64
-    cont->saved_thread.machine_register_stack_start = 0;
-    cont->saved_thread.machine_register_stack_end = 0;
+    sth->machine.register_stack_start = 0;
+    sth->machine.register_stack_end = 0;
 #endif
 }
 
@@ -445,7 +455,12 @@ cont_init(rb_context_t *cont, rb_thread_t *th)
 {
     /* save thread context */
     cont_save_thread(cont, th);
+    cont->saved_thread.self = th->self;
+    cont->saved_thread.machine.stack_maxsize = th->machine.stack_maxsize;
+    cont->saved_thread.fiber = th->fiber;
     cont->saved_thread.local_storage = 0;
+    cont->saved_thread.local_storage_recursive_hash = Qnil;
+    cont->saved_thread.local_storage_recursive_hash_for_trace = Qnil;
 }
 
 static rb_context_t *
@@ -464,19 +479,22 @@ cont_new(VALUE klass)
 
 static VALUE
 cont_capture(volatile int *stat)
+#if defined(__clang__) && \
+    __clang_major__ == 3 && __clang_minor__ == 8 && __clang_patch__ == 0
+__attribute__ ((optnone))
+#endif
 {
     rb_context_t *cont;
-    rb_thread_t *th = GET_THREAD(), *sth;
+    rb_thread_t *th = GET_THREAD();
     volatile VALUE contval;
 
     THREAD_MUST_BE_RUNNING(th);
     rb_vm_stack_to_heap(th);
     cont = cont_new(rb_cContinuation);
     contval = cont->self;
-    sth = &cont->saved_thread;
 
 #ifdef CAPTURE_JUST_VALID_VM_STACK
-    cont->vm_stack_slen = th->cfp->sp + th->mark_stack_len - th->stack;
+    cont->vm_stack_slen = th->cfp->sp - th->stack;
     cont->vm_stack_clen = th->stack + th->stack_size - (VALUE*)th->cfp;
     cont->vm_stack = ALLOC_N(VALUE, cont->vm_stack_slen + cont->vm_stack_clen);
     MEMCPY(cont->vm_stack, th->stack, VALUE, cont->vm_stack_slen);
@@ -485,7 +503,7 @@ cont_capture(volatile int *stat)
     cont->vm_stack = ALLOC_N(VALUE, th->stack_size);
     MEMCPY(cont->vm_stack, th->stack, VALUE, th->stack_size);
 #endif
-    sth->stack = 0;
+    cont->saved_thread.stack = 0;
 
     cont_save_machine_stack(th, cont);
 
@@ -508,6 +526,7 @@ cont_capture(volatile int *stat)
     if (ruby_setjmp(cont->jmpbuf)) {
 	volatile VALUE value;
 
+	VAR_INITIALIZED(cont);
 	value = cont->value;
 	if (cont->argc == -1) rb_exc_raise(value);
 	cont->value = Qnil;
@@ -520,7 +539,7 @@ cont_capture(volatile int *stat)
     }
 }
 
-static void
+static inline void
 cont_restore_thread(rb_context_t *cont)
 {
     rb_thread_t *th = GET_THREAD(), *sth = &cont->saved_thread;
@@ -528,16 +547,14 @@ cont_restore_thread(rb_context_t *cont)
     /* restore thread context */
     if (cont->type == CONTINUATION_CONTEXT) {
 	/* continuation */
-	VALUE fib;
+	rb_fiber_t *fib;
 
 	th->fiber = sth->fiber;
 	fib = th->fiber ? th->fiber : th->root_fiber;
 
 	if (fib) {
-	    rb_fiber_t *fcont;
-	    GetFiberPtr(fib, fcont);
-	    th->stack_size = fcont->cont.saved_thread.stack_size;
-	    th->stack = fcont->cont.saved_thread.stack;
+	    th->stack_size = fib->cont.saved_thread.stack_size;
+	    th->stack = fib->cont.saved_thread.stack;
 	}
 #ifdef CAPTURE_JUST_VALID_VM_STACK
 	MEMCPY(th->stack, cont->vm_stack, VALUE, cont->vm_stack_slen);
@@ -552,7 +569,9 @@ cont_restore_thread(rb_context_t *cont)
 	th->stack = sth->stack;
 	th->stack_size = sth->stack_size;
 	th->local_storage = sth->local_storage;
-	th->fiber = cont->self;
+	th->local_storage_recursive_hash = sth->local_storage_recursive_hash;
+	th->local_storage_recursive_hash_for_trace = sth->local_storage_recursive_hash_for_trace;
+	th->fiber = (rb_fiber_t*)cont;
     }
 
     th->cfp = sth->cfp;
@@ -579,7 +598,7 @@ fiber_set_stack_location(void)
     VALUE *ptr;
 
     SET_MACHINE_STACK_END(&ptr);
-    th->machine_stack_start = (void*)(((VALUE)ptr & RB_PAGE_MASK) + STACK_UPPER((void *)&ptr, 0, RB_PAGE_SIZE));
+    th->machine.stack_start = (void*)(((VALUE)ptr & RB_PAGE_MASK) + STACK_UPPER((void *)&ptr, 0, RB_PAGE_SIZE));
 }
 
 static VOID CALLBACK
@@ -654,7 +673,7 @@ fiber_initialize_machine_stack_context(rb_fiber_t *fib, size_t size)
 	    rb_raise(rb_eFiberError, "can't create fiber");
 	}
     }
-    sth->machine_stack_maxsize = size;
+    sth->machine.stack_maxsize = size;
 #else /* not WIN32 */
     ucontext_t *context = &fib->context;
     char *ptr;
@@ -665,12 +684,14 @@ fiber_initialize_machine_stack_context(rb_fiber_t *fib, size_t size)
     context->uc_link = NULL;
     context->uc_stack.ss_sp = ptr;
     context->uc_stack.ss_size = size;
+    fib->ss_sp = ptr;
+    fib->ss_size = size;
     makecontext(context, rb_fiber_start, 0);
-    sth->machine_stack_start = (VALUE*)(ptr + STACK_DIR_UPPER(0, size));
-    sth->machine_stack_maxsize = size - RB_PAGE_SIZE;
+    sth->machine.stack_start = (VALUE*)(ptr + STACK_DIR_UPPER(0, size));
+    sth->machine.stack_maxsize = size - RB_PAGE_SIZE;
 #endif
 #ifdef __ia64
-    sth->machine_register_stack_maxsize = sth->machine_stack_maxsize;
+    sth->machine.register_stack_maxsize = sth->machine.stack_maxsize;
 #endif
 }
 
@@ -687,31 +708,31 @@ fiber_setcontext(rb_fiber_t *newfib, rb_fiber_t *oldfib)
 
     /* restore thread context */
     cont_restore_thread(&newfib->cont);
-    th->machine_stack_maxsize = sth->machine_stack_maxsize;
-    if (sth->machine_stack_end && (newfib != oldfib)) {
-	rb_bug("fiber_setcontext: sth->machine_stack_end has non zero value");
+    th->machine.stack_maxsize = sth->machine.stack_maxsize;
+    if (sth->machine.stack_end && (newfib != oldfib)) {
+	rb_bug("fiber_setcontext: sth->machine.stack_end has non zero value");
     }
 
     /* save  oldfib's machine stack */
     if (oldfib->status != TERMINATED) {
 	STACK_GROW_DIR_DETECTION;
-	SET_MACHINE_STACK_END(&th->machine_stack_end);
+	SET_MACHINE_STACK_END(&th->machine.stack_end);
 	if (STACK_DIR_UPPER(0, 1)) {
-	    oldfib->cont.machine_stack_size = th->machine_stack_start - th->machine_stack_end;
-	    oldfib->cont.machine_stack = th->machine_stack_end;
+	    oldfib->cont.machine.stack_size = th->machine.stack_start - th->machine.stack_end;
+	    oldfib->cont.machine.stack = th->machine.stack_end;
 	}
 	else {
-	    oldfib->cont.machine_stack_size = th->machine_stack_end - th->machine_stack_start;
-	    oldfib->cont.machine_stack = th->machine_stack_start;
+	    oldfib->cont.machine.stack_size = th->machine.stack_end - th->machine.stack_start;
+	    oldfib->cont.machine.stack = th->machine.stack_start;
 	}
     }
     /* exchange machine_stack_start between oldfib and newfib */
-    oldfib->cont.saved_thread.machine_stack_start = th->machine_stack_start;
-    th->machine_stack_start = sth->machine_stack_start;
-    /* oldfib->machine_stack_end should be NULL */
-    oldfib->cont.saved_thread.machine_stack_end = 0;
+    oldfib->cont.saved_thread.machine.stack_start = th->machine.stack_start;
+    th->machine.stack_start = sth->machine.stack_start;
+    /* oldfib->machine.stack_end should be NULL */
+    oldfib->cont.saved_thread.machine.stack_end = 0;
 #ifndef _WIN32
-    if (!newfib->context.uc_stack.ss_sp && th->root_fiber != newfib->cont.self) {
+    if (!newfib->context.uc_stack.ss_sp && th->root_fiber != newfib) {
 	rb_bug("non_root_fiber->context.uc_stac.ss_sp should not be NULL");
     }
 #endif
@@ -742,16 +763,16 @@ cont_restore_1(rb_context_t *cont)
 	    ((_JUMP_BUFFER*)(&buf))->Frame;
     }
 #endif
-    if (cont->machine_stack_src) {
+    if (cont->machine.stack_src) {
 	FLUSH_REGISTER_WINDOWS;
-	MEMCPY(cont->machine_stack_src, cont->machine_stack,
-		VALUE, cont->machine_stack_size);
+	MEMCPY(cont->machine.stack_src, cont->machine.stack,
+		VALUE, cont->machine.stack_size);
     }
 
 #ifdef __ia64
-    if (cont->machine_register_stack_src) {
-	MEMCPY(cont->machine_register_stack_src, cont->machine_register_stack,
-	       VALUE, cont->machine_register_stack_size);
+    if (cont->machine.register_stack_src) {
+	MEMCPY(cont->machine.register_stack_src, cont->machine.register_stack,
+	       VALUE, cont->machine.register_stack_size);
     }
 #endif
 
@@ -786,7 +807,7 @@ register_stack_extend(rb_context_t *cont, VALUE *vp, VALUE *curr_bsp)
         E(k) = E(l) = E(m) = E(n) = E(o) =
         E(p) = E(q) = E(r) = E(s) = E(t) = 0;
     }
-    if (curr_bsp < cont->machine_register_stack_src+cont->machine_register_stack_size) {
+    if (curr_bsp < cont->machine.register_stack_src+cont->machine.register_stack_size) {
         register_stack_extend(cont, vp, (VALUE*)rb_ia64_bsp());
     }
     cont_restore_0(cont, vp);
@@ -798,7 +819,7 @@ register_stack_extend(rb_context_t *cont, VALUE *vp, VALUE *curr_bsp)
 static void
 cont_restore_0(rb_context_t *cont, VALUE *addr_in_prev_frame)
 {
-    if (cont->machine_stack_src) {
+    if (cont->machine.stack_src) {
 #ifdef HAVE_ALLOCA
 #define STACK_PAD_SIZE 1
 #else
@@ -811,7 +832,7 @@ cont_restore_0(rb_context_t *cont, VALUE *addr_in_prev_frame)
 	    /* Stack grows downward */
 #endif
 #if STACK_GROW_DIRECTION <= 0
-	    volatile VALUE *const end = cont->machine_stack_src;
+	    volatile VALUE *const end = cont->machine.stack_src;
 	    if (&space[0] > end) {
 # ifdef HAVE_ALLOCA
 		volatile VALUE *sp = ALLOCA_N(VALUE, &space[0] - end);
@@ -827,7 +848,7 @@ cont_restore_0(rb_context_t *cont, VALUE *addr_in_prev_frame)
 	    /* Stack grows upward */
 #endif
 #if STACK_GROW_DIRECTION >= 0
-	    volatile VALUE *const end = cont->machine_stack_src + cont->machine_stack_size;
+	    volatile VALUE *const end = cont->machine.stack_src + cont->machine.stack_size;
 	    if (&space[STACK_PAD_SIZE] < end) {
 # ifdef HAVE_ALLOCA
 		volatile VALUE *sp = ALLOCA_N(VALUE, end - &space[STACK_PAD_SIZE]);
@@ -927,7 +948,7 @@ rb_callcc(VALUE self)
 }
 
 static VALUE
-make_passing_arg(int argc, VALUE *argv)
+make_passing_arg(int argc, const VALUE *argv)
 {
     switch (argc) {
       case 0:
@@ -1043,9 +1064,6 @@ rb_cont_call(int argc, VALUE *argv, VALUE contval)
 	rb_raise(rb_eRuntimeError, "continuation called across stack rewinding barrier");
     }
     if (cont->saved_thread.fiber) {
-	rb_fiber_t *fcont;
-	GetFiberPtr(cont->saved_thread.fiber, fcont);
-
 	if (th->fiber != cont->saved_thread.fiber) {
 	    rb_raise(rb_eRuntimeError, "continuation called across fiber");
 	}
@@ -1079,7 +1097,7 @@ rb_cont_call(int argc, VALUE *argv, VALUE contval)
  *  comes with a small 4KB stack. This enables the fiber to be paused from deeply
  *  nested function calls within the fiber block.
  *
- *  When a fiber is created it will not run automatically. Rather it must be
+ *  When a fiber is created it will not run automatically. Rather it must
  *  be explicitly asked to run using the <code>Fiber#resume</code> method.
  *  The code running inside the fiber can give up control by calling
  *  <code>Fiber.yield</code> in which case it yields control back to caller
@@ -1131,7 +1149,7 @@ rb_cont_call(int argc, VALUE *argv, VALUE contval)
 static const rb_data_type_t fiber_data_type = {
     "fiber",
     {fiber_mark, fiber_free, fiber_memsize,},
-    NULL, NULL, RUBY_TYPED_FREE_IMMEDIATELY
+    0, 0, RUBY_TYPED_FREE_IMMEDIATELY
 };
 
 static VALUE
@@ -1151,12 +1169,11 @@ fiber_t_alloc(VALUE fibval)
     }
 
     THREAD_MUST_BE_RUNNING(th);
-    fib = ALLOC(rb_fiber_t);
-    memset(fib, 0, sizeof(rb_fiber_t));
+    fib = ZALLOC(rb_fiber_t);
     fib->cont.self = fibval;
     fib->cont.type = FIBER_CONTEXT;
     cont_init(&fib->cont, th);
-    fib->prev = Qnil;
+    fib->prev = NULL;
     fib->status = CREATED;
 
     DATA_PTR(fibval) = fib;
@@ -1170,6 +1187,7 @@ fiber_init(VALUE fibval, VALUE proc)
     rb_fiber_t *fib = fiber_t_alloc(fibval);
     rb_context_t *cont = &fib->cont;
     rb_thread_t *th = &cont->saved_thread;
+    rb_thread_t *cth = GET_THREAD();
 
     /* initialize cont */
     cont->vm_stack = 0;
@@ -1177,9 +1195,7 @@ fiber_init(VALUE fibval, VALUE proc)
     th->stack = 0;
     th->stack_size = 0;
 
-    fiber_link_join(fib);
-
-    th->stack_size = th->vm->default_params.fiber_vm_stack_size / sizeof(VALUE);
+    th->stack_size = cth->vm->default_params.fiber_vm_stack_size / sizeof(VALUE);
     th->stack = ALLOC_N(VALUE, th->stack_size);
 
     th->cfp = (void *)(th->stack + th->stack_size);
@@ -1200,11 +1216,13 @@ fiber_init(VALUE fibval, VALUE proc)
     th->cfp->me = 0;
     th->tag = 0;
     th->local_storage = st_init_numtable();
+    th->local_storage_recursive_hash = Qnil;
+    th->local_storage_recursive_hash_for_trace = Qnil;
 
     th->first_proc = proc;
 
 #if !FIBER_USE_NATIVE
-    MEMCPY(&cont->jmpbuf, &th->root_jmpbuf, rb_jmpbuf_t, 1);
+    MEMCPY(&cont->jmpbuf, &cth->root_jmpbuf, rb_jmpbuf_t, 1);
 #endif
 
     return fibval;
@@ -1223,61 +1241,19 @@ rb_fiber_new(VALUE (*func)(ANYARGS), VALUE obj)
     return fiber_init(fiber_alloc(rb_cFiber), rb_proc_new(func, obj));
 }
 
-static VALUE
-return_fiber(void)
-{
-    rb_fiber_t *fib;
-    VALUE curr = rb_fiber_current();
-    VALUE prev;
-    GetFiberPtr(curr, fib);
-
-    prev = fib->prev;
-    if (NIL_P(prev)) {
-	const VALUE root_fiber = GET_THREAD()->root_fiber;
-
-	if (root_fiber == curr) {
-	    rb_raise(rb_eFiberError, "can't yield from root fiber");
-	}
-	return root_fiber;
-    }
-    else {
-	fib->prev = Qnil;
-	return prev;
-    }
-}
-
-VALUE rb_fiber_transfer(VALUE fib, int argc, VALUE *argv);
-
-static void
-rb_fiber_terminate(rb_fiber_t *fib)
-{
-    VALUE value = fib->cont.value;
-    fib->status = TERMINATED;
-#if FIBER_USE_NATIVE && !defined(_WIN32)
-    /* Ruby must not switch to other thread until storing terminated_machine_stack */
-    terminated_machine_stack.ptr = fib->context.uc_stack.ss_sp;
-    terminated_machine_stack.size = fib->context.uc_stack.ss_size / sizeof(VALUE);
-    fib->context.uc_stack.ss_sp = NULL;
-    fib->cont.machine_stack = NULL;
-    fib->cont.machine_stack_size = 0;
-#endif
-    rb_fiber_transfer(return_fiber(), 1, &value);
-}
+static void rb_fiber_terminate(rb_fiber_t *fib);
 
 void
 rb_fiber_start(void)
 {
     rb_thread_t *th = GET_THREAD();
-    rb_fiber_t *fib;
-    rb_context_t *cont;
+    rb_fiber_t *fib = th->fiber;
     rb_proc_t *proc;
     int state;
 
-    GetFiberPtr(th->fiber, fib);
-    cont = &fib->cont;
-
     TH_PUSH_TAG(th);
     if ((state = EXEC_TAG()) == 0) {
+	rb_context_t *cont = &VAR_FROM_MEMORY(fib)->cont;
 	int argc;
 	const VALUE *argv, args = cont->value;
 	GetProcPtr(cont->saved_thread.first_proc, proc);
@@ -1321,90 +1297,117 @@ root_fiber_alloc(rb_thread_t *th)
 #endif
 #endif
     fib->status = RUNNING;
-    fib->prev_fiber = fib->next_fiber = fib;
 
     return fib;
 }
 
-VALUE
-rb_fiber_current(void)
+static inline rb_fiber_t*
+fiber_current(void)
 {
     rb_thread_t *th = GET_THREAD();
     if (th->fiber == 0) {
 	/* save root */
 	rb_fiber_t *fib = root_fiber_alloc(th);
-	th->root_fiber = th->fiber = fib->cont.self;
+	th->root_fiber = th->fiber = fib;
     }
     return th->fiber;
 }
 
-static VALUE
-fiber_store(rb_fiber_t *next_fib)
+static inline rb_fiber_t*
+return_fiber(void)
+{
+    rb_fiber_t *fib = fiber_current();
+    rb_fiber_t *prev = fib->prev;
+
+    if (!prev) {
+	rb_fiber_t *root_fiber = GET_THREAD()->root_fiber;
+
+	if (root_fiber == fib) {
+	    rb_raise(rb_eFiberError, "can't yield from root fiber");
+	}
+	return root_fiber;
+    }
+    else {
+	fib->prev = NULL;
+	return prev;
+    }
+}
+
+VALUE
+rb_fiber_current(void)
+{
+    return fiber_current()->cont.self;
+}
+
+static inline VALUE
+fiber_store(rb_fiber_t *next_fib, rb_thread_t *th)
 {
-    rb_thread_t *th = GET_THREAD();
     rb_fiber_t *fib;
 
     if (th->fiber) {
-	GetFiberPtr(th->fiber, fib);
+	fib = th->fiber;
 	cont_save_thread(&fib->cont, th);
     }
     else {
 	/* create current fiber */
 	fib = root_fiber_alloc(th);
-	th->root_fiber = th->fiber = fib->cont.self;
+	th->root_fiber = th->fiber = fib;
     }
 
-#if !FIBER_USE_NATIVE
-    cont_save_machine_stack(th, &fib->cont);
-#endif
-
-    if (FIBER_USE_NATIVE || ruby_setjmp(fib->cont.jmpbuf)) {
 #if FIBER_USE_NATIVE
-	fiber_setcontext(next_fib, fib);
+    fiber_setcontext(next_fib, fib);
+    /* restored */
 #ifndef _WIN32
-	if (terminated_machine_stack.ptr) {
-	    if (machine_stack_cache_index < MAX_MACHINE_STACK_CACHE) {
-		machine_stack_cache[machine_stack_cache_index].ptr = terminated_machine_stack.ptr;
-		machine_stack_cache[machine_stack_cache_index].size = terminated_machine_stack.size;
-		machine_stack_cache_index++;
+    if (terminated_machine_stack.ptr) {
+	if (machine_stack_cache_index < MAX_MACHINE_STACK_CACHE) {
+	    machine_stack_cache[machine_stack_cache_index].ptr = terminated_machine_stack.ptr;
+	    machine_stack_cache[machine_stack_cache_index].size = terminated_machine_stack.size;
+	    machine_stack_cache_index++;
+	}
+	else {
+	    if (terminated_machine_stack.ptr != fib->cont.machine.stack) {
+		munmap((void*)terminated_machine_stack.ptr, terminated_machine_stack.size * sizeof(VALUE));
 	    }
 	    else {
-		if (terminated_machine_stack.ptr != fib->cont.machine_stack) {
-		    munmap((void*)terminated_machine_stack.ptr, terminated_machine_stack.size * sizeof(VALUE));
-		}
-		else {
-		    rb_bug("terminated fiber resumed");
-		}
+		rb_bug("terminated fiber resumed");
 	    }
-	    terminated_machine_stack.ptr = NULL;
-	    terminated_machine_stack.size = 0;
 	}
-#endif
-#endif
+	terminated_machine_stack.ptr = NULL;
+	terminated_machine_stack.size = 0;
+    }
+#endif /* not _WIN32 */
+    fib = th->fiber;
+    if (fib->cont.argc == -1) rb_exc_raise(fib->cont.value);
+    return fib->cont.value;
+
+#else /* FIBER_USE_NATIVE */
+    cont_save_machine_stack(th, &fib->cont);
+    if (ruby_setjmp(fib->cont.jmpbuf)) {
 	/* restored */
-	GetFiberPtr(th->fiber, fib);
+	fib = th->fiber;
 	if (fib->cont.argc == -1) rb_exc_raise(fib->cont.value);
+	if (next_fib->cont.value == Qundef) {
+	    cont_restore_0(&next_fib->cont, &next_fib->cont.value);
+	    rb_bug("rb_fiber_resume: unreachable");
+	}
 	return fib->cont.value;
     }
-#if !FIBER_USE_NATIVE
     else {
-	return Qundef;
+	VALUE undef = Qundef;
+	cont_restore_0(&next_fib->cont, &undef);
+	rb_bug("rb_fiber_resume: unreachable");
     }
-#endif
+#endif /* FIBER_USE_NATIVE */
 }
 
 static inline VALUE
-fiber_switch(VALUE fibval, int argc, VALUE *argv, int is_resume)
+fiber_switch(rb_fiber_t *fib, int argc, const VALUE *argv, int is_resume)
 {
     VALUE value;
-    rb_fiber_t *fib;
-    rb_context_t *cont;
+    rb_context_t *cont = &fib->cont;
     rb_thread_t *th = GET_THREAD();
 
-    GetFiberPtr(fibval, fib);
-    cont = &fib->cont;
-
-    if (th->fiber == fibval) {
+    if (th->fiber == fib) {
 	/* ignore fiber context switch
          * because destination fiber is same as current fiber
 	 */
@@ -1419,34 +1422,25 @@ fiber_switch(VALUE fibval, int argc, VALUE *argv, int is_resume)
     }
     else if (fib->status == TERMINATED) {
 	value = rb_exc_new2(rb_eFiberError, "dead fiber called");
-	if (th->fiber != fibval) {
-	    GetFiberPtr(th->fiber, fib);
-	    if (fib->status != TERMINATED) rb_exc_raise(value);
-	    fibval = th->root_fiber;
-	}
-	else {
-	    fibval = fib->prev;
-	    if (NIL_P(fibval)) fibval = th->root_fiber;
-	}
-	GetFiberPtr(fibval, fib);
-	cont = &fib->cont;
+
+	if (th->fiber->status != TERMINATED) rb_exc_raise(value);
+
+	/* th->fiber is also dead => switch to root fiber */
+	/* (this means we're being called from rb_fiber_terminate, */
+	/* and the terminated fiber's return_fiber() is already dead) */
+	cont = &th->root_fiber->cont;
 	cont->argc = -1;
 	cont->value = value;
 #if FIBER_USE_NATIVE
-	{
-	    VALUE oldfibval;
-	    rb_fiber_t *oldfib;
-	    oldfibval = rb_fiber_current();
-	    GetFiberPtr(oldfibval, oldfib);
-	    fiber_setcontext(fib, oldfib);
-	}
+	fiber_setcontext(th->root_fiber, th->fiber);
 #else
 	cont_restore_0(cont, &value);
 #endif
+	/* unreachable */
     }
 
     if (is_resume) {
-	fib->prev = rb_fiber_current();
+	fib->prev = fiber_current();
     }
     else {
 	/* restore `tracing' context. see [Feature #4347] */
@@ -1456,56 +1450,67 @@ fiber_switch(VALUE fibval, int argc, VALUE *argv, int is_resume)
     cont->argc = argc;
     cont->value = make_passing_arg(argc, argv);
 
-    value = fiber_store(fib);
-#if !FIBER_USE_NATIVE
-    if (value == Qundef) {
-	cont_restore_0(cont, &value);
-	rb_bug("rb_fiber_resume: unreachable");
-    }
-#endif
+    value = fiber_store(fib, th);
     RUBY_VM_CHECK_INTS(th);
 
     return value;
 }
 
 VALUE
-rb_fiber_transfer(VALUE fib, int argc, VALUE *argv)
+rb_fiber_transfer(VALUE fibval, int argc, const VALUE *argv)
 {
+    rb_fiber_t *fib;
+    GetFiberPtr(fibval, fib);
     return fiber_switch(fib, argc, argv, 0);
 }
 
+static void
+rb_fiber_terminate(rb_fiber_t *fib)
+{
+    VALUE value = fib->cont.value;
+    fib->status = TERMINATED;
+#if FIBER_USE_NATIVE && !defined(_WIN32)
+    /* Ruby must not switch to other thread until storing terminated_machine_stack */
+    terminated_machine_stack.ptr = fib->ss_sp;
+    terminated_machine_stack.size = fib->ss_size / sizeof(VALUE);
+    fib->ss_sp = NULL;
+    fib->context.uc_stack.ss_sp = NULL;
+    fib->cont.machine.stack = NULL;
+    fib->cont.machine.stack_size = 0;
+#endif
+    fiber_switch(return_fiber(), 1, &value, 0);
+}
+
 VALUE
-rb_fiber_resume(VALUE fibval, int argc, VALUE *argv)
+rb_fiber_resume(VALUE fibval, int argc, const VALUE *argv)
 {
     rb_fiber_t *fib;
     GetFiberPtr(fibval, fib);
 
-    if (fib->prev != Qnil || fib->cont.type == ROOT_FIBER_CONTEXT) {
+    if (fib->prev != 0 || fib->cont.type == ROOT_FIBER_CONTEXT) {
 	rb_raise(rb_eFiberError, "double resume");
     }
-    if (fib->transfered != 0) {
+    if (fib->transferred != 0) {
 	rb_raise(rb_eFiberError, "cannot resume transferred Fiber");
     }
 
-    return fiber_switch(fibval, argc, argv, 1);
+    return fiber_switch(fib, argc, argv, 1);
 }
 
 VALUE
-rb_fiber_yield(int argc, VALUE *argv)
+rb_fiber_yield(int argc, const VALUE *argv)
 {
-    return rb_fiber_transfer(return_fiber(), argc, argv);
+    return fiber_switch(return_fiber(), argc, argv, 0);
 }
 
 void
 rb_fiber_reset_root_local_storage(VALUE thval)
 {
     rb_thread_t *th;
-    rb_fiber_t	*fib;
 
     GetThreadPtr(thval, th);
     if (th->root_fiber && th->root_fiber != th->fiber) {
-	GetFiberPtr(th->root_fiber, fib);
-	th->local_storage = fib->cont.saved_thread.local_storage;
+	th->local_storage = th->root_fiber->cont.saved_thread.local_storage;
     }
 }
 
@@ -1585,7 +1590,7 @@ rb_fiber_m_resume(int argc, VALUE *argv, VALUE fib)
  *    fiber2.resume
  *    fiber3.resume
  *
- *    <em>produces</em>
+ *  <em>produces</em>
  *
  *    In fiber 2
  *    In fiber 1
@@ -1597,8 +1602,8 @@ rb_fiber_m_transfer(int argc, VALUE *argv, VALUE fibval)
 {
     rb_fiber_t *fib;
     GetFiberPtr(fibval, fib);
-    fib->transfered = 1;
-    return rb_fiber_transfer(fibval, argc, argv);
+    fib->transferred = 1;
+    return fiber_switch(fib, argc, argv, 0);
 }
 
 /*
@@ -1659,7 +1664,7 @@ Init_Cont(void)
 #else /* not WIN32 */
     pagesize = sysconf(_SC_PAGESIZE);
 #endif
-    SET_MACHINE_STACK_END(&th->machine_stack_end);
+    SET_MACHINE_STACK_END(&th->machine.stack_end);
 #endif
 
     rb_cFiber = rb_define_class("Fiber", rb_cObject);
